/* (C) 2022 YiRing, Inc. */
package com.yiring.app.service.message.impl;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.yiring.app.domain.district.District;
import com.yiring.app.domain.district.DistrictRepository;
import com.yiring.app.domain.key.*;
import com.yiring.app.domain.location.*;
import com.yiring.app.domain.log.ZyRealtimeLog;
import com.yiring.app.domain.log.ZyRealtimeLogRepository;
import com.yiring.app.param.key.KeyAlarmAddParam;
import com.yiring.app.service.message.PositionMessageService;
import com.yiring.app.stomp.message.PositionMessage;
import com.yiring.app.util.GeoUtils;
import com.yiring.app.vo.equipment.EquipmentVo;
import com.yiring.auth.domain.dept.Department;
import com.yiring.auth.domain.user.User;
import com.yiring.auth.domain.user.UserRepository;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import javax.annotation.Resource;
import javax.persistence.criteria.Expression;
import javax.persistence.criteria.Predicate;
import lombok.extern.slf4j.Slf4j;
import org.locationtech.jts.geom.Point;
import org.springframework.beans.BeanUtils;
import org.springframework.data.domain.*;
import org.springframework.messaging.simp.SimpMessagingTemplate;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

/**
 * 真源消息处理实现类
 *
 * @author Jim
 * @version 0.1
 * 2022/5/9 10:22
 */

@Slf4j
@Service
@Transactional(rollbackFor = Exception.class)
public class PositionMessageServiceImpl implements PositionMessageService {

    @Resource
    LocationTagRepository locationTagRepository;

    @Resource
    LocationLogRepository locationLogRepository;

    @Resource
    LocationFenceRepository locationFenceRepository;

    @Resource
    DistrictRepository districtRepository;

    @Resource
    LocationTurnoverRepository locationTurnoverRepository;

    @Resource
    SimpMessagingTemplate simpMessagingTemplate;

    @Resource
    ZyRealtimeLogRepository zyRealtimeLogRepository;

    @Resource
    KeyAlarmRepository keyAlarmRepository;

    @Resource
    KeyAlarmLogRepository keyAlarmLogRepository;

    @Resource
    KeyAlarmRuleRepository keyAlarmRuleRepository;

    @Resource
    UserRepository userRepository;

    @Override
    public void consume(String message) {
        // 将消息转换成 JSON 格式
        JSONObject info = JSON.parseObject(message);
        log.debug("Receiver Message: {}", info);

        // 解构消息内容
        JSONObject data = info.getJSONObject("params");
        String method = info.getString("method");
        if (method == null) {
            log.warn("Unknown Message: {}", info);
            return;
        }

        // 记录日志
        ZyRealtimeLog realtimeLog = ZyRealtimeLog.builder().method(method).raw(info).build();
        zyRealtimeLogRepository.save(realtimeLog);

        // 业务处理
        switch (method) {
            // 实时定位
            case "position" -> processPositionMessage(data);
            // 设备低电量
            case "lowPower" -> processLowPowerMessage(data);
            // 设备状态变更
            case "deviceStatus" -> processDeviceStatusMessage(data);
            // 按键报警
            case "keyWarning" -> processKeyWarningMessage(data);
            // 围栏报警
            case "enclosure" -> log.warn("Ignore Message Type [enclosure]: {}", info);
            default -> log.warn("Unknown Message Type: {}", info);
        }
    }

    /**
     * 处理定位消息
     * @param data 消息内容
     */
    public void processPositionMessage(JSONObject data) {
        // TODO
        log.info("Position Message: {}", data);

        // 查询定位标签
        String tagId = data.getString("tagId");
        Example<LocationTag> example = Example.of(LocationTag.builder().code(tagId).build());
        Optional<LocationTag> optional = locationTagRepository.findOne(example);
        if (optional.isEmpty()) {
            log.warn("Tag Not Found: {}", tagId);
            return;
        }

        // 构建复合主键
        LocationTag tag = optional.get();
        TagTimeId id = new TagTimeId(tag);

        // 定位时间
        Instant instant = Instant.ofEpochMilli(data.getLongValue("locationTime"));
        LocalDateTime locationTime = LocalDateTime.ofInstant(instant, ZoneId.systemDefault());

        // 定位的基本信息
        LocationLog locationLog = LocationLog
            .builder()
            .id(id)
            .raw(data)
            .locationTime(locationTime)
            .areaId(data.getLong("areaId"))
            .floor(data.getString("floor"))
            .silent(data.getBoolean("silent"))
            .volt(data.getInteger("volt"))
            .voltUnit(data.getString("voltUnit"))
            .out(data.getBoolean("out"))
            .build();

        // 获取定位卡当前绑定的用户
        User user = tag.getUser();
        if (user != null) {
            // 定位标签当时所属的用户
            locationLog.setUser(user);
            // 查询当前用户的状态
            locationLog.setStatus(user.getStatus());
        }

        // 设置空间点位信息
        Point point = GeoUtils.createPoint(
            data.getDoubleValue("longitude"),
            data.getDoubleValue("latitude"),
            data.getDoubleValue("altitude")
        );
        locationLog.setPoint(point);

        // 定位信标
        String beacons = data.getString("beacons");
        if (beacons != null) {
            Set<String> beaconCodes = Arrays
                .stream(beacons.split(","))
                .map(beacon -> beacon.replaceAll("\\(.*\\)", ""))
                .collect(Collectors.toSet());
            locationLog.setBeacons(new JSONArray().fluentAddAll(beaconCodes));
        }

        // 查询所有围栏信息
        List<LocationFence> fences = locationFenceRepository.findAll();
        // 找出当前定位在哪些围栏内（同楼层，经纬度范围内）
        Set<Long> fenceIds = fences
            .stream()
            .filter(fence -> fence.getMapId().equals(locationLog.getAreaId()) && fence.getGeometry().contains(point))
            .map(LocationFence::getId)
            .collect(Collectors.toSet());
        locationLog.setFences(new JSONArray().fluentAddAll(fenceIds));

        // 查询出所有的区域
        List<District> districts = districtRepository.findAll();
        // 找出当前定位在哪些围栏内（同楼层，经纬度范围内）
        Set<Long> districtIds = districts
            .stream()
            .filter(district -> point.within(district.getGeometry()))
            .map(District::getId)
            .collect(Collectors.toSet());
        locationLog.setDistricts(new JSONArray().fluentAddAll(districtIds));

        // 找出区域与围栏配置的最大防抖时间
        Optional<LocationFence> maxThresholdFence = fences
            .stream()
            .max(Comparator.comparing(LocationFence::getThreshold));
        Optional<District> maxThresholdDistrict = districts
            .stream()
            .max(Comparator.comparing(District::getDebouncingDuration));
        Optional<Integer> maxThreshold = Stream
            .of(
                maxThresholdFence.map(LocationFence::getThreshold).orElse(0),
                maxThresholdDistrict.map(District::getDebouncingDuration).orElse(0)
            )
            .max(Integer::compareTo);
        if (maxThreshold.get() > 0) {
            // 查询最大防抖时间段内的定位日志记录
            List<LocationLog> logs = findByTagTimeIdAndThreshold(id, maxThreshold.get());

            // 计算出入标记（围栏、区域）
            List<LocationTurnover> turnovers = new ArrayList<>();
            // 统计所有围栏
            turnovers.addAll(statisticalTurnovers(fences, id, logs));
            // 统计所有区域
            turnovers.addAll(statisticalTurnovers(districts, id, logs));
            // 写入围栏/区域进出记录
            if (!turnovers.isEmpty()) {
                locationTurnoverRepository.saveAll(turnovers);
            }
        }

        // 写入定位数据
        locationLogRepository.saveAndFlush(locationLog);

        // 更新定位标签卡状态信息
        if (!tag.getSilent().equals(locationLog.getSilent())) {
            // 如果运动状态发生变更，记录状态变更时间
            tag.setSilentStatusUpdateTime(id.getTime());
        }
        tag.setSilent(locationLog.getSilent());
        tag.setOut(locationLog.getOut());
        tag.setPoint(locationLog.getPoint());
        tag.setVolt(locationLog.getVolt());
        tag.setVoltUnit(locationLog.getVoltUnit());
        locationTagRepository.save(tag);

        // 更新围栏内的标签数据
        locationFenceRepository.saveAll(fences);
        // 更新区域内的标签数据
        districtRepository.saveAll(districts);

        // WebSocket 推送定位消息
        // 消息内容需要确定 TODO
        PositionMessage message = PositionMessage
            .builder()
            .time(id.getTime())
            .tagId(id.getTag().getId())
            .tagCode(id.getTag().getCode())
            .point(locationLog.getPoint())
            .build();
        simpMessagingTemplate.convertAndSend("/topic/position", message);
    }

    /**
     * 统计进出记录
     * @param areas 区域/围栏列表
     * @param id 定位日志主键
     * @param logs 定位日志列表
     * @return 进出记录列表
     */
    public <E> List<LocationTurnover> statisticalTurnovers(List<E> areas, TagTimeId id, List<LocationLog> logs) {
        List<LocationTurnover> turnovers = new ArrayList<>();

        // 遍历所有区域/围栏
        for (Object area : areas) {
            // 收集不同表共同关键信息
            LocationTurnover.Type type;
            Long sourceId;
            int threshold;
            if (area instanceof LocationFence fence) {
                sourceId = fence.getId();
                threshold = fence.getThreshold();
                type = LocationTurnover.Type.FENCE;
            } else if (area instanceof District district) {
                sourceId = district.getId();
                threshold = district.getDebouncingDuration();
                type = LocationTurnover.Type.DISTRICT;
            } else {
                continue;
            }

            // 获取需要过滤的时间段
            LocalDateTime startTime = id.getTime().minusSeconds(threshold);
            LocalDateTime endTime = id.getTime();

            // 过滤出围栏防抖时间段内的定位日志记录
            Set<LocationLog> collect = logs
                .stream()
                .filter(item -> item.getId().getTime().isBefore(endTime) && item.getId().getTime().isAfter(startTime))
                .collect(Collectors.toSet());
            // 找出定位记录中围栏集合
            LocationTurnover.Type finalType = type;
            List<JSONArray> includes = collect
                .stream()
                .map(item -> LocationTurnover.Type.FENCE.equals(finalType) ? item.getFences() : item.getDistricts())
                .toList();
            if (includes.isEmpty()) {
                continue;
            }

            // 检查是否进入区域
            Boolean isEnter = checkEnter(includes, sourceId);
            if (isEnter != null) {
                // 检查是否为重复进入/退出（一定是先有进入记录才会有离开记录）
                Optional<LocationTurnover> repeat = findRepeatTurnoverRecord(sourceId, type, id.getTag(), isEnter);
                if (repeat.isEmpty()) {
                    // 尝试将上一次记录标记为非最新记录
                    trySetPrevTurnoverExpired(sourceId, type, id.getTag());

                    // 记录当前标签进入/离开围栏/区域的动作，标记为最新记录
                    LocationTurnover turnover = LocationTurnover
                        .builder()
                        .enter(isEnter)
                        .type(type)
                        .sourceId(sourceId)
                        .tag(id.getTag())
                        .time(id.getTime())
                        .user(id.getTag().getUser())
                        .isLatest(true)
                        .build();
                    turnovers.add(turnover);

                    // 更新范围内的标签信息
                    if (area instanceof LocationFence fence) {
                        fence.setTags(resetInTags(id.getTag(), isEnter, fence.getTags()));
                    } else {
                        District district = (District) area;
                        district.setTags(resetInTags(id.getTag(), isEnter, district.getTags()));
                    }
                }
            }
        }

        return turnovers;
    }

    /**
     * 根据定位日志 ID 和防抖时间查询定位日志
     * @param id 定位日志 ID
     * @param threshold 防抖时间
     * @return 定位日志集合
     */
    public List<LocationLog> findByTagTimeIdAndThreshold(TagTimeId id, int threshold) {
        return locationLogRepository.findAll((root, query, cb) -> {
            Predicate predicate = cb.conjunction();
            List<Expression<Boolean>> expressions = predicate.getExpressions();
            expressions.add(cb.equal(root.get("id").get("tag"), id.getTag()));
            expressions.add(cb.greaterThanOrEqualTo(root.get("id").get("time"), id.getTime().minusSeconds(threshold)));
            expressions.add(cb.lessThanOrEqualTo(root.get("id").get("time"), id.getTime()));
            return predicate;
        });
    }

    /**
     * 根据防抖期间进入的区域/围栏以及新产生的围栏/区域记录，判断是否进入
     * @param array 区域/围栏集合
     * @param id 区域/围栏 ID
     * @return true 进入，false 退出，null 未发生变化
     */
    public Boolean checkEnter(List<JSONArray> array, Long id) {
        Boolean isEnter = null;
        long count = array.stream().filter(ids -> ids != null && ids.contains(id)).count();
        if (count == array.size()) {
            isEnter = true;
        } else if (count == 0) {
            isEnter = false;
        }
        return isEnter;
    }

    /**
     * 查找
     * @param sourceId 区域/围栏 ID
     * @param type 类型（区域/围栏）
     * @param tag 定位标签
     * @param enter true 进入，false 退出
     * @return true 重复进入，false 非重复进入
     */
    public Optional<LocationTurnover> findRepeatTurnoverRecord(
        Long sourceId,
        LocationTurnover.Type type,
        LocationTag tag,
        Boolean enter
    ) {
        Pageable pageable = PageRequest.of(0, 1, Sort.by(Sort.Order.desc(LocationTurnover.Fields.time)));
        Page<LocationTurnover> page = locationTurnoverRepository.findAll(
            (root, query, cb) -> {
                Predicate predicate = cb.conjunction();
                List<Expression<Boolean>> expressions = predicate.getExpressions();
                expressions.add(cb.equal(root.get(LocationTurnover.Fields.sourceId), sourceId));
                expressions.add(cb.equal(root.get(LocationTurnover.Fields.type), type));
                expressions.add(cb.equal(root.get(LocationTurnover.Fields.tag), tag));
                return predicate;
            },
            pageable
        );

        long elements = page.getTotalElements();
        if (elements == 0 && Boolean.FALSE.equals(enter)) {
            // 如果标签在目标区域没有过进入记录，但是确判断为离开，这时需拟一个记录，避免记录一个没有进入确有离开记录的情况
            return Optional.of(new LocationTurnover());
        }

        if (elements > 0) {
            // 检查是否重复进入
            if (enter == null || page.get().anyMatch(turnover -> turnover.getEnter() == enter)) {
                return page.stream().findFirst();
            }
        }

        // 需要记录
        return Optional.empty();
    }

    /**
     * 尝试设置上一个进出记录为过期标识
     * @param sourceId 区域/围栏 ID
     * @param type 类型（区域/围栏）
     * @param tag 定位标签
     */
    public void trySetPrevTurnoverExpired(Long sourceId, LocationTurnover.Type type, LocationTag tag) {
        Optional<LocationTurnover> record = findRepeatTurnoverRecord(sourceId, type, tag, null);
        if (record.isPresent()) {
            LocationTurnover turnoverRecord = record.get();
            turnoverRecord.setIsLatest(false);
            locationTurnoverRepository.save(turnoverRecord);
        }
    }

    /**
     * 重置标签集合
     * @param tag 标签
     * @param enter 进入/离开
     * @param tags 标签集合
     * @return 重置后的标签集合
     */
    public Set<LocationTag> resetInTags(LocationTag tag, Boolean enter, Set<LocationTag> tags) {
        if (Boolean.TRUE.equals(enter)) {
            tags.add(tag);
        } else {
            tags.remove(tag);
        }
        return tags;
    }

    /**
     * 处理低电量报警消息
     * @param data 消息内容
     */
    public void processLowPowerMessage(JSONObject data) {
        // TODO
        log.info("LowPower Message: {}", data);
    }

    /**
     * 处理设备状态更新消息
     * @param data 消息内容
     */
    public void processDeviceStatusMessage(JSONObject data) {
        // TODO
        log.info("DeviceStatus Message: {}", data);
        EquipmentVo equipmentVo = data.toJavaObject(EquipmentVo.class);
        if (equipmentVo.getDeviceType().equals("BTI")) {
            Optional<LocationTag> one = locationTagRepository.findOne(
                Example.of(LocationTag.builder().code(equipmentVo.getDeviceId()).build())
            );
            if (one.isEmpty()) {
                log.error("没找到该标签");
                return;
            }
            //修改标签电量
            LocationTag locationTag = LocationTag
                .builder()
                .code(equipmentVo.getDeviceId())
                .volt(equipmentVo.getVolt())
                .voltUnit(equipmentVo.getField_21())
                .build();
            BeanUtils.copyProperties(one.get(), locationTag);
            locationTagRepository.saveAndFlush(locationTag);
        }
    }

    /**
     * 处理按键报警消息
     * @param data 消息内容
     */
    public void processKeyWarningMessage(JSONObject data) {
        // TODO
        log.info("KeyWarning Message: {}", data);
        KeyAlarmAddParam keyAlarmAddParam = data.toJavaObject(KeyAlarmAddParam.class);
        KeyAlarm keyAlarm = KeyAlarmAddParam.transform(keyAlarmAddParam);
        keyAlarm.setPoint(
            GeoUtils.xyzToPoint(
                keyAlarm.getX().doubleValue(),
                keyAlarm.getY().doubleValue(),
                keyAlarm.getZ().doubleValue()
            )
        );

        Optional<User> userRepositoryOne = userRepository.findOne(
            Example.of(User.builder().linkId(keyAlarmAddParam.getEntityId()).build())
        );

        if (userRepositoryOne.isEmpty()) {
            //            throw new RuntimeException("没找到该用户");
            log.warn("Tag Not Found: {}", "没找到该用户");
            return;
        }

        User user = userRepositoryOne.get();

        keyAlarm.setLeader(user);
        keyAlarm.setEnable(true);
        Optional<KeyAlarmRule> keyAlarmRuleOptional = keyAlarmRuleRepository.findOne(
            Example.of(
                KeyAlarmRule.builder().department(Department.builder().id(user.getDepartment().getId()).build()).build()
            )
        );

        if (keyAlarmRuleOptional.isEmpty()) {
            //            throw new RuntimeException("该部门未设置报警规则");
            log.warn("Tag Not Found: {}", "该部门未设置报警规则");
            return;
        }

        Optional<LocationTag> locationTagOptional = locationTagRepository.findOne(
            Example.of(LocationTag.builder().code(keyAlarmAddParam.getTagId()).build())
        );

        if (locationTagOptional.isEmpty()) {
            log.warn("Tag Not Found: {}", keyAlarmAddParam.getTagId());
            return;
        }
        keyAlarm.setCode(locationTagOptional.get().getCode());

        final KeyAlarm finalKeyAlarm = keyAlarmRepository.save(keyAlarm);

        KeyAlarmRule keyAlarmRule = keyAlarmRuleOptional.get();
        Set<User> users = keyAlarmRule.getUsers();

        HashSet<KeyAlarmLog> keyAlarmLogs = new HashSet<>();
        users.forEach(leader -> {
            String[] split = keyAlarmRule.getTypes().split(",");
            for (String type : split) {
                keyAlarmLogs.add(
                    KeyAlarmLog.builder().status(false).user(leader).types(type).keyAlarm(finalKeyAlarm).build()
                );
            }
        });

        keyAlarmLogRepository.saveAll(keyAlarmLogs);
    }
}
